
import collections

from itertools import product
from typing import Sequence, Dict, Any

from coba.registry import CobaRegistry
from coba.pipes import Pipe, Filter, MemorySource

from coba.benchmarks.core import Benchmark

class BenchmarkFileFmtV2(Filter[Dict[str,Any], Benchmark]):

    def filter(self, config: Dict[str,Any]) -> 'Benchmark':

        variables = { k: CobaRegistry.construct(v) for k,v in config.get("variables",{}).items() }

        def _construct(item:Any) -> Sequence[Any]:
            result = None

            if isinstance(item, str) and item in variables:
                result = variables[item]

            if isinstance(item, str) and item not in variables:
                result = CobaRegistry.construct(item)

            if isinstance(item, dict):
                result = CobaRegistry.construct(item)

            if isinstance(item, list):
                if any([ isinstance(i,list) for i in item ]):
                    raise Exception("Recursive structures are not supported in benchmark simulation configs.")
                pieces = list(map(_construct, item))
                result = [ Pipe.join(s, f) for s in pieces[0] for f in product(*pieces[1:])]

            if result is None:
                raise Exception(f"We were unable to construct {item} in the given benchmark file.")

            return result if isinstance(result, collections.Sequence) else [result]

        if not isinstance(config['simulations'], list): config['simulations'] = [config['simulations']]

        simulations = [ simulation for recipe in config['simulations'] for simulation in _construct(recipe)]

        return Benchmark(simulations)



from copy import deepcopy
from itertools import groupby, product, count
from collections import defaultdict
from typing import Iterable, Sequence, Any, Optional, Dict, Hashable, Tuple

from coba.random import CobaRandom
from coba.learners import Learner, SafeLearner
from coba.config import CobaConfig
from coba.utilities import PackageChecker
from coba.pipes import Source, Pipe, Filter, IdentityFilter
from coba.simulations import Simulation, Interaction, Shuffle, Take, OpenmlSimulation, ClassificationSimulation

from coba.benchmarks.transactions import Transaction
from coba.benchmarks.results import Result

class Identifier():
    
    def __init__(self) -> None:
        self._source_ids   : Dict[Hashable, int]  = defaultdict(lambda x=count(): next(x)) # type: ignore
        self._learner_ids  : Dict[Hashable, int]  = defaultdict(lambda x=count(): next(x)) # type: ignore
        self._simulaion_ids: Dict[Hashable, int]  = defaultdict(lambda x=count(): next(x)) # type: ignore

    def id(self, simulation: Simulation, learner: Learner) -> Tuple[int,int,int]:
        source = simulation._source if isinstance(simulation, Pipe.SourceFilters) else simulation
        
        src_id = self._source_ids[source]
        sim_id = self._simulaion_ids[simulation]
        lrn_id = self._learner_ids[learner] if learner else None
        
        return (src_id, sim_id, lrn_id)

class Task(Filter[Iterable[Interaction], Iterable[Any]]):

    def __init__(self, src_id:int, sim_id: int, lrn_id: int, simulation: Simulation, learner: Optional[Learner]) -> None:

        self.sim_pipe   = simulation
        self.sim_source = simulation._source if isinstance(simulation, Pipe.SourceFilters) else simulation
        self.sim_filter = simulation._filter if isinstance(simulation, Pipe.SourceFilters) else IdentityFilter()
        self.learner    = SafeLearner(learner) if learner else None

        self.src_id = src_id
        self.sim_id = sim_id
        self.lrn_id = lrn_id

class EvaluationTask(Task):

    def __init__(self, src_id:int, sim_id: int, lrn_id: int, simulation: Simulation, learner: Learner, seed: int) -> None:
        self._seed = seed
        super().__init__(src_id, sim_id, lrn_id, simulation, learner)

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Any]:

        if not interactions: return

        learner = deepcopy(self.learner)
        random  = CobaRandom(self._seed)

        with CobaConfig.Logger.time(f"Evaluating learner {self.lrn_id} on Simulation {self.sim_id}..."):

            row_data = defaultdict(list)

            for i, interaction in enumerate(interactions):

                context   = interaction.context
                actions   = interaction.actions
                feedbacks = interaction.feedbacks

                probs  = learner.predict(i, context, actions)

                assert abs(sum(probs) - 1) < .0001, "The learner returned invalid proabilities for action choices."

                action = random.choice(actions, probs)
                reward = feedbacks[actions.index(action)]
                prob   = probs[actions.index(action)]

                info = learner.learn(i, context, action, reward, prob) or {}
                                                        
                for key,value in info.items() | {('reward',reward)}: 
                    row_data[key].append(value)

            yield Transaction.interactions(self.sim_id, self.lrn_id, _packed=row_data)

class SimulationTask(Task):

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Any]:

        with CobaConfig.Logger.time(f"Calculating Simulation {self.sim_id} statistics..."):
            extra_statistics = {}

            contexts,actions,feedbacks = zip(*[ (i.context, i.actions, i.feedbacks) for i in interactions])

            if isinstance(self.sim_source, (ClassificationSimulation,OpenmlSimulation)):

                try:
                    PackageChecker.sklearn("")

                    from sklearn.feature_extraction import FeatureHasher
                    from sklearn.ensemble import RandomForestClassifier
                    from sklearn.model_selection import cross_val_score

                    X   = contexts
                    y   = [ a[f.index(1)] for a,f in zip(actions,feedbacks)]
                    clf = RandomForestClassifier(n_estimators=50)

                    if any(isinstance(f,str) for f in X[0]):
                        X = [ dict(enumerate(x)) for x in X ]

                    if isinstance(X[0],dict):
                        X = [ dict(zip(map(str,x.keys()), x.values())) for x in X ]
                        X = FeatureHasher(n_features=2**17).fit_transform(X)

                    if len(y) > 5:
                        extra_statistics["bayes_rate"] = round(cross_val_score(clf, X, y, cv=5).mean(),4)

                except ImportError:
                    pass

                labels     = set()
                features   = set() 
                label_cnts = defaultdict(int)

                for c,a,f in zip(contexts,actions,feedbacks):

                    inter_label = a[f.index(1)]
                    inter_feats = c.keys() if isinstance(c,dict) else range(len(c))

                    labels.add(inter_label)
                    features.update(inter_feats)
                    label_cnts[inter_label] += 1

                extra_statistics["action_cardinality"] = len(labels)
                extra_statistics["context_dimensions"] = len(features)
                extra_statistics["imbalance_ratio"]    = round(max(label_cnts.values())/min(label_cnts.values()),4)

            if isinstance(self.sim_filter,Pipe.FiltersFilter):
                filters = self.sim_filter._filters
            elif isinstance(self.sim_filter, IdentityFilter):
                filters = []
            else:
                filters = [self.sim_filter]

            source  = str(self.sim_source).strip('"')
            shuffle = "None"
            take    = "None"
            pipe    = str(self.sim_pipe) 

            for filter in filters:
                if isinstance(filter, Shuffle): shuffle = str(filter._seed )
                if isinstance(filter, Take   ): take    = str(filter._count)

            yield Transaction.simulation(self.sim_id, 
                source=source, 
                shuffle=shuffle, 
                take=take, 
                pipe=pipe, 
                **extra_statistics
            )

class CreateTasks(Source[Iterable[Task]]):

    def __init__(self, simulations: Sequence[Simulation], learners: Sequence[Learner], seed: int = None) -> None:
        self._simulations = simulations
        self._learners    = learners
        self._seed        = seed

    def read(self) -> Iterable[Task]:

        #we rely on sim_id to make sure we don't do duplicate work. So long as self._simulations
        #is always in the exact same order we should be fine. In the future we may want to consider.
        #adding a better check for simulations other than assigning an index based on their order.

        identifier = Identifier()

        for simulation in self._simulations:
            yield SimulationTask(*identifier.id(simulation, None), simulation, None)

        for simulation, learner in product(self._simulations, self._learners):
            yield EvaluationTask(*identifier.id(simulation, learner), simulation, learner, self._seed)

class FilterFinished(Filter[Iterable[Task], Iterable[Task]]):
    def __init__(self, restored: Result) -> None:
        self._restored = restored

    def filter(self, tasks: Iterable[Task]) -> Iterable[Task]:

        def is_not_complete(task: Task):

            if isinstance(task,SimulationTask):
                return task.sim_id not in self._restored.simulations

            if isinstance(task,EvaluationTask):
                return (task.sim_id,task.lrn_id) not in self._restored._interactions

            raise Exception("Unrecognized Task")

        return filter(is_not_complete, tasks)

class ChunkBySource(Filter[Iterable[Task], Iterable[Iterable[Task]]]):

    def filter(self, tasks: Iterable[Task]) -> Iterable[Iterable[Task]]:

        srt_key = lambda t: t.src_id
        grp_key = lambda t: t.src_id

        tasks = list(tasks)

        for _, group in groupby(sorted(tasks, key=srt_key), key=grp_key):
            yield list(group)

class ChunkByTask(Filter[Iterable[Task], Iterable[Iterable[Task]]]):

    def filter(self, tasks: Iterable[Task]) -> Iterable[Iterable[Task]]:

        for task in tasks:
            yield [ task ]

class ChunkByNone(Filter[Iterable[Task], Iterable[Iterable[Task]]]):

    def filter(self, tasks: Iterable[Task]) -> Iterable[Iterable[Task]]:
        yield list(tasks)

class ProcessTasks(Filter[Iterable[Iterable[Task]], Iterable[Any]]):

    def filter(self, chunks: Iterable[Iterable[Task]]) -> Iterable[Any]:

        for chunk in chunks:
            for transaction in self._process_chunk(chunk):
                yield transaction

    def _process_chunk(self, task_group: Iterable[Task]) -> Iterable[Any]:

        source_by_id = { t.src_id: t.sim_source for t in task_group }
        filter_by_id = { t.sim_id: t.sim_filter for t in task_group }

        srt_src = lambda t: t.src_id
        grp_src = lambda t: t.src_id
        srt_sim = lambda t: t.sim_id
        grp_sim = lambda t: t.sim_id

        with CobaConfig.Logger.log(f"Processing chunk..."):

            for src_id, tasks_by_src in groupby(sorted(task_group, key=srt_src), key=grp_src):

                try:

                    with CobaConfig.Logger.time(f"Creating source {src_id} from {source_by_id[src_id]}..."):
                        #This is not ideal. I'm not sure how it should be improved so it is being left for now.
                        loaded_source = list(source_by_id[src_id].read())

                    for sim_id, tasks_by_src_sim in groupby(sorted(tasks_by_src, key=srt_sim), key=grp_sim):

                        with CobaConfig.Logger.time(f"Creating simulation {sim_id} from source {src_id}..."):
                            interactions = filter_by_id[sim_id].filter(loaded_source)

                        if not interactions:
                            CobaConfig.Logger.log(f"Simulation {sim_id} has nothing to evaluate (likely due to `take` being larger than the simulation).")
                            return

                        for task in tasks_by_src_sim:
                            try:
                                for transaction in task.filter(interactions): 
                                    yield transaction
                            except Exception as e:
                                CobaConfig.Logger.log_exception(e)

                except Exception as e:
                    CobaConfig.Logger.log_exception(e)


from coba.simulations.core import Interaction
from pathlib import Path
from itertools import product
from typing import Iterable, Sequence, cast, Optional, overload, List, Union

from coba.learners import Learner
from coba.simulations import Simulation, Take, Shuffle
from coba.registry import CobaRegistry
from coba.config import CobaConfig, CobaFatal
from coba.pipes import Pipe, Filter, Source, JsonDecode, ResponseToLines, HttpSource, MemorySource, DiskSource
from coba.multiprocessing import MultiprocessFilter

from coba.benchmarks.tasks import ChunkByNone, CreateTasks, FilterFinished, ChunkByTask, ChunkBySource, ProcessTasks
from coba.benchmarks.transactions import Transaction, TransactionSink
from coba.benchmarks.results import Result

class Benchmark:
    """A Benchmark which uses simulations to calculate performance statistics for learners."""
    
    @overload
    @staticmethod
    def from_file(filesource:Union[Source[str], Source[Iterable[str]]]) -> 'Benchmark': ...

    @overload
    @staticmethod
    def from_file(filename:str) -> 'Benchmark': ...
    
    @staticmethod #type: ignore #(this apppears to be a mypy bug https://github.com/python/mypy/issues/7781)
    def from_file(arg) -> 'Benchmark': #type: ignore
        """Instantiate a Benchmark from a config file."""

        if isinstance(arg,str) and arg.startswith('http'):
            content = '\n'.join(ResponseToLines().filter(HttpSource(arg).read()))
        
        elif isinstance(arg,str) and not arg.startswith('http'):
            content = '\n'.join(DiskSource(arg).read())

        else:
            content = arg.read() #type: ignore

        return CobaRegistry.construct(CobaConfig.Benchmark['file_fmt']).filter(JsonDecode().filter(content))

    def __init__(self, 
        simulations: Sequence[Simulation],
        shuffle    : Sequence[Optional[int]] = [None],
        take       : int = None) -> None:
        """Instantiate a Benchmark.

        Args:
            simulations: The collection of simulations to benchmark against.
            shuffle: A collection of seeds to use for simulation shuffling. A seed of `None` means no shuffle will be applied.
            take: The number of interactions to take from each simulation for evaluation.
        """
        ...

        sources: List[Simulation] = simulations
        filters: List[Sequence[Filter[Iterable[Interaction],Iterable[Interaction]]]] = []

        if shuffle != [None]:
            filters.append([ Shuffle(seed) for seed in shuffle ])

        if take is not None:
            filters.append([ Take(take) ])

        if len(filters) > 0:
            simulation_sources = [cast(Simulation,Pipe.join(s,f)) for s,f in product(sources, product(*filters))]
        else:
            simulation_sources = list(sources)

        self._simulations         : Sequence[Simulation] = simulation_sources
        self._processes           : Optional[int]        = None
        self._maxtasksperchild    : Optional[int]        = None
        self._maxtasksperchild_set: bool                 = False
        self._chunk_by            : Optional[str]        = None

    def chunk_by(self, value: str = 'source') -> 'Benchmark':
        """Determines how tasks are chunked for processing.
        
        Args:
            value: Allowable values are 'task', 'source' and 'none'.
        """

        assert value in ['task', 'source', 'none'], "The given chunk_by value wasn't recognized. Allowed values are 'task', 'source' and 'none'"

        self._chunk_by = value

        return self

    def processes(self, value:int = 1) -> 'Benchmark':
        """Determines how many processes will be utilized for processing Benchmark chunks.
        
        Args:
            value: This is the number of processes Benchmark will use.
        """

        self._processes = value
        return self

    def maxtasksperchild(self, value: Optional[int] = None) -> 'Benchmark':
        """Determines how many chunks a process can handle before it will be torn down and recreated.
        
        Args:
            value: This is the number of chunks a process will handle before being recreated. If this
                value is None then processes will remain alive for the life of the Benchmark evaluation.
        """

        self._maxtasksperchild_set = True
        self._maxtasksperchild = value
        return self

    def evaluate(self, learners: Sequence[Learner], result_file:str = None, seed:int = 1) -> Result:
        """Collect observations of a Learner playing the benchmark's simulations to calculate Results.

        Args:
            learners: The collection of learners that we'd like to evalute.
            result_file: The file we'd like to use for writing/restoring results for the requested evaluation.
            seed: The random seed we'd like to use when choosing which action to take from the learner's predictions.

        Returns:
            See the base class for more information.
        """

        restored = Result.from_file(result_file) if result_file and Path(result_file).exists() else Result()

        n_given_learners    = len(learners)
        n_given_simulations = len(self._simulations)
 
        if len(restored.benchmark) != 0:
            assert n_given_learners    == restored.benchmark['n_learners'   ], "The currently evaluating benchmark doesn't match the given transaction log"
            assert n_given_simulations == restored.benchmark['n_simulations'], "The currently evaluating benchmark doesn't match the given transaction log"

        preamble = []
        preamble.append(Transaction.version())
        preamble.append(Transaction.benchmark(n_given_learners, n_given_simulations))
        preamble.extend(Transaction.learners(learners))

        cb = self._chunk_by         if self._chunk_by             else CobaConfig.Benchmark['chunk_by']
        mp = self._processes        if self._processes            else CobaConfig.Benchmark['processes']
        mt = self._maxtasksperchild if self._maxtasksperchild_set else CobaConfig.Benchmark['maxtasksperchild']
            
        tasks            = CreateTasks(self._simulations, learners, seed)
        unfinished       = FilterFinished(restored)
        chunked          = ChunkByTask() if cb == 'task' else ChunkByNone() if cb == 'none' else ChunkBySource()
        process          = ProcessTasks()
        transaction_sink = TransactionSink(result_file, restored)

        if mp > 1 or mt is not None  : process = MultiprocessFilter([process], mp, mt) #type: ignore

        try:
            Pipe.join(MemorySource(preamble), []                            , transaction_sink).run()
            Pipe.join(tasks                 , [unfinished, chunked, process], transaction_sink).run()
        except KeyboardInterrupt:
            CobaConfig.Logger.log("Benchmark evaluation was manually aborted via Ctrl-C")
        except CobaFatal:
            raise
        except Exception as ex:
            CobaConfig.Logger.log_exception(ex)

        return transaction_sink.result